from builtins import *
import os
import sys
import numpy as np

from contextlib import contextmanager

def incrementCount(start):
    count = start
    while True:
        yield count
        count += 1

class _silentWriter(object):
    def write(self, message):
        pass

class Silence(object):
    '''
    Context manager to silence any output except
    when using the returned override function.

    Usage:

    >> with Silence() as override:
    >>     print(1)
    >>     with override():
    >>         print(2)
    >>     print(3)

    2
    '''
    #https://stackoverflow.com/questions/8391411/suppress-calls-to-print-python
    #https://stackoverflow.com/questions/2828953/silence-the-stdout-of-a-function-in-python-without-trashing-sys-stdout-and-resto

    @contextmanager
    def _override(self):
        sys.stdout = self._original_stdout
        yield
        sys.stdout = _silentWriter()

    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = _silentWriter()
        return self._override

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout = self._original_stdout



import os

PACKAGE_PATH = os.path.dirname(__file__)
RESOURCE_PATH = os.path.abspath(os.path.join(PACKAGE_PATH, 'resources'))
CHAIN_PATH = os.path.abspath(os.path.join(RESOURCE_PATH, 'chains'))
DEFAULT_CONFIG_PATH = os.path.join(RESOURCE_PATH, 'config/defaultConfig.txt')
TEST_CONFIG_PATH = os.path.join(RESOURCE_PATH, 'config/testComboConfig.txt')

from blendz.config import Configuration
from blendz.photoz import Photoz
from blendz.model import BPZ, LikelihoodOnly



from builtins import *
from os.path import join
import numpy as np
from blendz import Configuration

class Filters(object):
    """Definitions of the photometric filter bands observations have been made with.

    More description of filters would presumably go here.

    Args:
        config (blendz.Configuratio): Description of parameter `config`. Defaults to None.

        kwargs : Description of parameter `kwargs`.

    Returns:
        type: Description of returned object.

    """
    def __init__(self, config=None, **kwargs):
        self.config = Configuration(**kwargs)
        if config is not None:
            self.config.mergeFromOther(config)

        self.filter_path = self.config.filter_path
        self.filter_names = self.config.filters
        self.num_filters = len(self.filter_names)

        self._all_filters = {}
        #load_filters reads filters in from file *and* calculates the normalisations
        self.load_filters()

    def load_filters(self, filenames=None, filepath=None):
        #Default arguments evaluated at define, self only available at function call, so use None instead
        if filenames is None:
            filenames = self.filter_names
        if filepath is None:
            filepath = self.filter_path

        for F in range(len(filenames)):
            #Read from file
            self._all_filters[F] = {}
            self._all_filters[F]['lambda'], self._all_filters[F]['response'] = \
                    np.loadtxt(join(filepath, filenames[F]), unpack=True)
            flt_order = np.argsort(self._all_filters[F]['lambda'])
            self._all_filters[F]['lambda'] = self._all_filters[F]['lambda'][flt_order]
            self._all_filters[F]['response'] = self._all_filters[F]['response'][flt_order]
            #Calculate normalisation
            self._all_filters[F]['norm'] = np.trapz(self._all_filters[F]['response'] / self._all_filters[F]['lambda'],\
                                            x = self._all_filters[F]['lambda'])

    def wavelength(self, F):
        try:
            return self._all_filters[F]['lambda']
        except (KeyError, TypeError):
            raise ValueError('Filter may be an integer [0...{}], but got a {} of value {} instead'.format(self.num_filters-1, type(F), F))

    def response(self, F):
        try:
            return self._all_filters[F]['response']
        except (KeyError, TypeError):
            raise ValueError('Filter may be an integer [0...{}], but got a {} of value {} instead'.format(self.num_filters-1, type(F), F))

    def norm(self, F):
        try:
            return self._all_filters[F]['norm']
        except (KeyError, TypeError):
            raise ValueError('Filter may be an integer [0...{}], but got a {} of value {} instead'.format(self.num_filters-1, type(F), F))



from builtins import *
import numpy as np


class Galaxy(object):
    def __init__(self, mag_data, mag_sigma, config, zero_point_frac, index):
        self.mag_data = mag_data
        self.mag_sigma = mag_sigma

        self.config = config
        self.ref_band = self.config.ref_band
        self.ref_mag_data = self.mag_data[self.ref_band]
        self.ref_mag_sigma = self.mag_sigma[self.ref_band]
        self.zero_point_frac = zero_point_frac
        self.index = index

        self.truth = {}

        self.convertMagnitudes()

    def convertMagnitudes(self):
        self.flux_data = 10.**(-0.4*self.mag_data)
        self.flux_sigma = (10.**(0.4*self.mag_sigma)-1.)*self.flux_data

        # Galaxy not OBSERVED - set flux to zero and error to very high (1e108)
        noObs = np.where(np.isclose(self.mag_data, self.config.no_observe_value, atol=0.01))[0]
        for no in noObs:
            self.flux_data[no] = 0.
            self.flux_sigma[no] = 1e108

        # Galaxy not DETECTED - set flux to zero, error to abs(error)
        noDet = np.where(np.isclose(self.mag_data, self.config.no_detect_value, atol=0.01))[0]
        for nd in noDet:
            self.flux_data[nd] = 0.
            self.flux_sigma[nd] = 10.**(-0.4*self.mag_sigma[nd])

        #Array of bools for bands where galaxy is neither not-observed nor not-detected
        # i.e., the galaxy IS seen in this band
        notSeen = np.array([((b in noObs) or (b in noDet)) for b in range(len(self.mag_data))])
        seen = np.array([not b for b in notSeen])

        if not seen[self.ref_band]:
            print(self.mag_data)
            raise ValueError('Galaxies must be observed in reference band, but galaxy {} had ref-band magnitude of {}, i.e., marked as a non-observation/detection.'.format(self.index, self.ref_mag_data))

        #Add the zero point errors in
        #First, handle the observed objects
        self.flux_sigma = np.where(seen, np.sqrt(self.flux_sigma*self.flux_sigma+(self.zero_point_frac*self.flux_data)**2.), self.flux_sigma)
        #Handle objects not seen differently to observed objects
        self.flux_sigma = np.where(notSeen, np.sqrt(self.flux_sigma*self.flux_sigma+(self.zero_point_frac*(self.flux_sigma/2.))**2), self.flux_sigma)

        #Create attribute for the flux in the reference band
        self.ref_flux_data = self.flux_data[self.config.ref_band]
        self.ref_flux_sigma = self.flux_sigma[self.config.ref_band]

        #Flux data and sigma, with the reference band removed
        self.flux_data_noRef = self.flux_data[self.config.non_ref_bands]
        self.flux_sigma_noRef = self.flux_sigma[self.config.non_ref_bands]



from builtins import *
import warnings
import numpy as np
from blendz import Configuration
from blendz.photometry import PhotometryBase, Galaxy

class Photometry(PhotometryBase):
    def __init__(self, config=None, **kwargs):
        super(Photometry, self).__init__(config=config, **kwargs)
        self.data_path = self.config.data_path
        self.zero_point_errors = self.config.zero_point_errors
        if self.config.data_is_csv:
            self.photo_data = np.loadtxt(self.data_path,
                skiprows=self.config.skip_data_rows, delimiter=',')
        else:
            self.photo_data = np.loadtxt(self.data_path,
                skiprows=self.config.skip_data_rows)
        self.num_to_load = np.shape(self.photo_data)[0]
        self.zero_point_frac = 10.**(0.4*self.zero_point_errors) - 1.

        self.loadGalaxies()

    def loadGalaxies(self):
        for g in range(self.num_to_load):
            mag_data = self.photo_data[g, self.config.mag_cols]
            mag_sigma = self.photo_data[g, self.config.sigma_cols]
            self.all_galaxies.append(Galaxy(mag_data, mag_sigma, self.config, self.zero_point_frac, g))
            if self.config.spec_z_col is not None:
                self.all_galaxies[g].truth['num_components'] = len(self.config.spec_z_col)
                for c in range(len(self.config.spec_z_col)):
                    cmp_redshift = self.photo_data[g, self.config.spec_z_col[c]]
                    self.all_galaxies[g].truth[c] = {'redshift': cmp_redshift}
                    if not (self.config.z_lo <= cmp_redshift <= self.config.z_hi):
                        warn_str = 'Galaxy {} has a spectroscopic redshift of {} in component {}. '.format(g, cmp_redshift, c) \
                                 + 'This is outside of the prior range ({} -> {}).'.format(self.config.z_lo, self.config.z_hi)
                        warnings.warn(warn_str)

            if self.config.magnitude_limit_col is not None:
                self.all_galaxies[g].magnitude_limit = self.photo_data[g, self.config.magnitude_limit_col]
            else:
                self.all_galaxies[g].magnitude_limit = self.config.magnitude_limit

            #If ref_mag_hi_sigma is not set, use ref_mag_hi instead, but prefer sigma
            if self.config.ref_mag_hi_sigma is None:
                if self.config.ref_mag_hi is not None:
                    self.all_galaxies[g].ref_mag_hi = self.config.ref_mag_hi
                else:
                    raise ValueError('One of ref_mag_hi or ref_mag_hi_sigma must be set in config to load photometry.')
            else:
                ref_flux_hi = self.config.ref_mag_hi_sigma * self.all_galaxies[g].ref_flux_sigma
                ref_mag_hi = -2.5 * np.log10(ref_flux_hi)
                self.all_galaxies[g].ref_mag_hi = ref_mag_hi



from builtins import *
import warnings
from contextlib import contextmanager
import numpy as np
from blendz import Configuration

class PhotometryBase(object):
    def __init__(self, config=None, **kwargs):
        self.all_galaxies = []
        self.current_galaxy = None

        self.config = Configuration(**kwargs)
        if config is not None:
            self.config.mergeFromOther(config)

    def iterate(self, start=None, stop=None, step=None):
        out_list = self.all_galaxies[start:stop:step]
        for gal in out_list:
            self.current_galaxy = gal
            yield gal
        #Clean up by resetting current_galaxy to None when done
        self.current_galaxy = None

    def __iter__(self):
        iterator = self.iterate()
        for g in range(len(self.all_galaxies)):
            yield next(iterator)
        #Clean up by resetting current_galaxy to None when done
        self.current_galaxy = None

    def __getitem__(self, key):
        out = self.all_galaxies[key]
        if isinstance(out, list):
            warnings.warn('This slice of the photometry returns a list of '
                          + 'Galaxy objects, but doesn\'t update current_galaxy, '
                          + 'so this should not be used for iterating over if any '
                          + 'methods are called. Instead, you should use the '
                          + 'iterate(start, stop, step) method for iterating.')
        return out

    @contextmanager
    def galaxy(self, index):
        gal = self.all_galaxies[index]
        self.current_galaxy = gal
        yield gal
        self.current_galaxy = None

    @property
    def num_galaxies(self):
        '''
        Read only attribute returning the number of galaxies contained in the photometry data set.
        '''
        return len(self.all_galaxies)



from builtins import *
from os.path import join, abspath, dirname
import warnings
try:
    #Python 2
    import ConfigParser
except ImportError:
    #Python 3
    import configparser as ConfigParser
import numpy as np
import blendz

class DefaultConfiguration(object):
    def __init__(self):
        self.configs_to_read = [blendz.DEFAULT_CONFIG_PATH]
        self.kwargs = {}

    def populate_configuration(self):
        self.readConfigFiles()
        self.saveAndConvertValues()

    def readConfigFiles(self):
        '''
        Set up the configparser and read the configuration file from disk. The
        default file is read first then (optionally) overwritten by a user
        supplied configuration.
        '''
        self.config = ConfigParser.SafeConfigParser()
        #Add the resourse_path to the ConfigParser so that it can be referenced in the config files
        self.config.set('DEFAULT', 'resource_path', blendz.RESOURCE_PATH)
        self.config.read(self.configs_to_read)

    def maybeGet(self, section, key, typeFn):
        '''
        Return a setting `key` from section `section`, where `typeFn` converts
        a string into the desired type. Settings are checked for first in the
        keyword arguments to the Configuration class, and then read from
        configuration files.
        '''
        if key in self.kwargs:
            return typeFn(self.kwargs[key])
        #Put special behaviour for if typeFn is bool, as bool() on any
        #non-empty string returns True, even if that string is "False"
        elif typeFn==bool:
            return self.config.getboolean(section, key)
        else:
            return typeFn(self.config.get(section, key))

    def maybeGetList(self, section, key, typeFn):
        '''
        Return a setting `key` from section `section`, where the setting is a
        list values of a single type, and `typeFn` converts a string into that
        desired type. Settings are checked for first in the keyword arguments
        to the Configuration class, and then read from configuration files.
        '''
        if key in self.kwargs:
            if self.kwargs[key] is None:
                return None
            else:
                try:
                    return [typeFn(v) for v in self.kwargs[key]]
                except (TypeError, ValueError):
                    return [typeFn(self.kwargs[key])]
        else:
            try:
                return [typeFn(v.strip()) for v in self.config.get(section, key).split(',')]
            except (TypeError, ValueError):
                return None

    def saveAndConvertValues(self):
        #Run config
        # We reserve np.nan for unset parameters that are ALWAYS allowed to be overwritten when merging!
        try:
            self._z_lo = self.maybeGet('Run', 'z_lo', float)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self._z_hi = self.maybeGet('Run', 'z_hi', float)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self._z_len = self.maybeGet('Run', 'z_len', int)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self._template_set = self.maybeGet('Run', 'template_set', str)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self._template_set_path = self.maybeGet('Run', 'template_set_path', str)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self.sort_redshifts = self.maybeGet('Run', 'sort_redshifts', bool)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self.ref_mag_hi = self.maybeGet('Run', 'ref_mag_hi', float)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            self.ref_mag_hi = None

        try:
            self.ref_mag_hi_sigma = self.maybeGet('Run', 'ref_mag_hi_sigma', float)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            self.ref_mag_hi_sigma = None

        try:
            self.ref_mag_lo = self.maybeGet('Run', 'ref_mag_lo', float)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self.omega_mat = self.maybeGet('Run', 'omega_mat', float)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self.omega_lam = self.maybeGet('Run', 'omega_lam', float)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self.omega_k = self.maybeGet('Run', 'omega_k', float)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self.hubble = self.maybeGet('Run', 'hubble', float)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self.r0 = self.maybeGet('Run', 'r0', float)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self.gamma = self.maybeGet('Run', 'gamma', float)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self.xi_r_cutoff = self.maybeGet('Run', 'xi_r_cutoff', float)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self.model_type = self.maybeGet('Run', 'model_type', str)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        # _prior_params is either an array of parameters,
        # or np.nan for when we want to do calibration (allow overwrite)
        try:
            _prior_params = self.maybeGetList('Run', 'prior_params', float)
            if _prior_params is None:
                self.prior_params = np.nan
            else:
                self.prior_params = np.array(_prior_params)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        #Data config
        try:
            self.data_path = self.maybeGet('Data', 'data_path', str)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self.skip_data_rows = self.maybeGet('Data', 'skip_data_rows', int)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self.data_is_csv = self.maybeGet('Data', 'data_is_csv', bool)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            #self.mag_cols = [int(i) for i in self.maybeGet('Data', 'mag_cols').split(',')]
            self.mag_cols = self.maybeGetList('Data', 'mag_cols', int)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            #self.sigma_cols = [int(i) for i in self.maybeGet('Data', 'sigma_cols').split(',')]
            self.sigma_cols = self.maybeGetList('Data', 'sigma_cols', int)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self._ref_band = self.maybeGet('Data', 'ref_band', int)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        #spec_z_col is never allowed to not be set, force it to None if not set
        #reading None can cause exceptions too, so catch them and set to None
        try:
            self.spec_z_col = self.maybeGetList('Data', 'spec_z_col', int)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            self.spec_z_col = None

        try:
            self.magnitude_limit_col = self.maybeGet('Data', 'magnitude_limit_col', int)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            self.magnitude_limit_col = None

        try:
            self.magnitude_limit = self.maybeGet('Data', 'magnitude_limit', float)
            if self.magnitude_limit_col is not None:
                warnings.warn('Both magnitude_limit and magnitude_limit_col '
                              + 'have been set, so magnitude_limit will be ignored.')
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            self.magnitude_limit = None

        try:
            self.filter_path = self.maybeGet('Data', 'filter_path', str)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self.filters = self.maybeGetList('Data', 'filters', str)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self.zero_point_errors = np.array(self.maybeGetList('Data', 'zero_point_errors', float))
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self.no_detect_value = self.maybeGet('Data', 'no_detect_value', float)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        try:
            self.no_observe_value = self.maybeGet('Data', 'no_observe_value', float)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

        #Angle within which we assume the galaxies lie, given they're blended
        #For use in correlation functions, given in arcsec, converted to rads here
        try:
            res_arcsec = self.maybeGet('Data', 'angular_resolution', float)
            self.angular_resolution = (res_arcsec / (60. * 60.)) * (np.pi / 180.)
        except (ConfigParser.NoOptionError, ConfigParser.NoSectionError):
            pass

    #Derived attiribute -> property for ref_band and non_ref_bands indices
    @property #getter
    def ref_band(self):
        return self._ref_band
    @ref_band.setter
    def ref_band(self, value):
        self.recalculate_non_ref_bands = True
        self._ref_band = value

    @property #getter, no setter so read-only
    def non_ref_bands(self):
        try:
            recalc = self.recalculate_non_ref_bands
        except AttributeError:
            recalc = True
        if recalc:
            self._non_ref_bands = np.ones(len(self.filters), dtype=bool)
            self._non_ref_bands[self.ref_band] = False
            self.recalculate_non_ref_bands = False
        return self._non_ref_bands

    #Make the attributes controlling the redshift_grid properties so we can
    #detect if they have changed and mark redshift_grid as needing recalculation
    @property #getter
    def z_lo(self):
        return self._z_lo
    @z_lo.setter
    def z_lo(self, value):
        self.recalculate_redshift_grid = True
        self._z_lo = value

    @property #getter
    def z_hi(self):
        return self._z_hi
    @z_hi.setter
    def z_hi(self, value):
        self.recalculate_redshift_grid = True
        self._z_hi = value

    @property #getter
    def z_len(self):
        return self._z_len
    @z_len.setter
    def z_len(self, value):
        self.recalculate_redshift_grid = True
        self._z_len = value

    @property #getter, no setter so read-only
    def redshift_grid(self):
        try:
            recalc = self.recalculate_redshift_grid
        except AttributeError:
            recalc = True
        if recalc:
            self._redshift_grid = np.linspace(self.z_lo, self.z_hi, self.z_len)
            self.recalculate_redshift_grid = False
        return self._redshift_grid

    #Do the same thing for the template_dict
    #For the template_dict, main config points to an info file, which
    #is itself a configuration file, containing the path and type of each template
    @property #getter
    def template_set(self):
        return self._template_set
    @template_set.setter
    def template_set(self, value):
        self.recalculate_template_dict = True
        self._template_set = value

    @property #getter
    def template_set_path(self):
        return self._template_set_path
    @template_set_path.setter
    def template_set_path(self, value):
        self.recalculate_template_dict = True
        self._template_set_path = value

    @property #getter, no setter so read-only
    def template_dict(self):
        try:
            recalc = self.recalculate_template_dict
        except AttributeError:
            recalc = True
        if recalc:
            #Use the path and name of the template set to load in the templates
            #if we need to do it again (because one of those has changed)
            template_config = ConfigParser.SafeConfigParser()
            template_config.read(join(self.template_set_path, self.template_set))
            self._template_dict = {}
            for i, template_name in enumerate(template_config.sections()):
                rel_path_t = template_config.get(template_name, 'path')
                abs_path_t = join(self.template_set_path, rel_path_t)
                type_t = template_config.get(template_name, 'type')
                self._template_dict[i] = {'name':template_name, 'path':abs_path_t, 'type':type_t}
            #Mark as not needing recalculating (unless of the properties changes)
            self.recalculate_template_dict = False
        return self._template_dict

    def __eq__(self, other):
        '''
        Check for equality of configurations. This will return True if all the settings are the
        same in both configurations, even if they are different objects.
        '''
        no_check = ['config', 'configs_to_read', 'kwargs']
        if self.__class__ != other.__class__:
            return False
        else:
            all_true = True
            for key in self.__dict__:
                if key not in no_check:
                    try:
                        all_true *= np.all(self.__dict__[key] == other.__dict__[key])
                    except KeyError:
                        #KeyError means a setting in self is not in other, so we
                        #know that the configs aren't equal.
                        all_true = False
            for key in other.__dict__:
                if key not in no_check:
                    try:
                        all_true *= np.all(self.__dict__[key] == other.__dict__[key])
                    except KeyError:
                        #KeyError means a setting in self is not in other, so we
                        #know that the configs aren't equal.
                        all_true = False
            return all_true

    def __ne__(self, other):
        #https://stackoverflow.com/a/30676267
        return not self == other



from builtins import *
from os.path import join
import warnings
import numpy as np
from scipy.interpolate import interp1d
from blendz import Configuration

class Templates(object):
    def __init__(self, config=None, **kwargs):
        self.config = Configuration(**kwargs)
        if config is not None:
            self.config.mergeFromOther(config)

        self.template_dict = self.config.template_dict
        self.num_templates = len(self.template_dict)
        self.possible_types = sorted(set(tmp['type'] for tmp in self.template_dict.values()))

        self.loadTemplates()
        self._num_type = self._countTypes()
        self._interpolators = self._getInterpolators()

        self.tmp_ind_to_type_ind = []
        for T in range(self.num_templates):
            tmpType = self.templateType(T)
            self.tmp_ind_to_type_ind.append(np.where(np.array(self.possible_types)==tmpType)[0][0])


    def loadTemplates(self):
        self._all_templates = {}
        for T in range(self.num_templates):
            self._all_templates[T] = {}
            self._all_templates[T]['lambda'], self._all_templates[T]['flux'] = \
                        np.loadtxt(self.template_dict[T]['path'], unpack=True)
            tmp_order = np.argsort(self._all_templates[T]['lambda'])
            self._all_templates[T]['lambda'] = self._all_templates[T]['lambda'][tmp_order]
            self._all_templates[T]['flux'] = self._all_templates[T]['flux'][tmp_order]
            self._all_templates[T]['name'] = self.template_dict[T]['name']

    def _countTypes(self):
        type_dict = {}
        for tmpType in self.possible_types:
            type_dict[tmpType] = len([tmp['type'] for tmp in self.template_dict.values()\
                                      if tmp['type']==tmpType])
        return type_dict

    def _getInterpolators(self):
        interpolators = {}
        for T in range(self.num_templates):
            interpolators[T] = interp1d(self.wavelength(T), self.flux(T), bounds_error=False, fill_value=0.)
        return interpolators

    def numType(self, tmpType):
        try:
            return self._num_type[tmpType]
        except KeyError:
            warnings.warn('There are no templates of type {}, returning zero.'.format(tmpType))
            return 0
        except TypeError:
            raise TypeError('A value of type {} cannot be a template-type.'.format(type(tmpType)))

    def templateType(self, T):
        try:
            return self.template_dict[T]['type']
        except (KeyError, TypeError):
            raise ValueError('Template may be an integer [0...{}], but got a {} of value {} instead'.format(self.num_templates-1, type(T), T))

    def wavelength(self, T):
        try:
            return self._all_templates[T]['lambda']
        except (KeyError, TypeError):
            raise ValueError('Template may be an integer [0...{}], but got a {} of value {} instead'.format(self.num_templates-1, type(T), T))

    def flux(self, T):
        try:
            return self._all_templates[T]['flux']
        except (KeyError, TypeError):
            raise ValueError('Template may be an integer [0...{}], but got a {} of value {} instead'.format(self.num_templates-1, type(T), T))

    def name(self, T):
        try:
            return self._all_templates[T]['name']
        except (KeyError, TypeError):
            raise ValueError('Template may be an integer [0...{}], but got a {} of value {} instead'.format(self.num_templates-1, type(T), T))

    def interp(self, T, new_lambda):
        try:
            return self._interpolators[T](new_lambda)
        except (KeyError, TypeError):
            raise ValueError('Template may be an integer [0...{}], but got a {} of value {} instead'.format(self.num_templates-1, type(T), T))



from builtins import *
import numpy as np
from scipy.interpolate import interp1d
from scipy.optimize import minimize
from scipy.special import logsumexp
from itertools import repeat, combinations
from blendz.model import ModelBase

class BPZ(ModelBase):
    #def __init__(self, mag_grid_len=100, max_ref_mag_hi=None, **kwargs):
    def __init__(self, mag_grid_len=100, **kwargs):
        super(BPZ, self).__init__(**kwargs)
        self.mag_grid_len = mag_grid_len

    @property #getter, no setter so read-only
    def prior_params_dict(self):
        try:
            # Return the precalculated value
            return self._prior_params_dict
        except AttributeError:
            # First time, precalculate and return if parameters are set
            if (self.prior_params is not np.nan):
                self._loadParameterDict()
                return self._prior_params_dict
            else:
                # Complain if they're not
                raise ValueError('Trying to use prior without setting prior parameters. '
                                 + 'Either set in config file or run calibration.')

    @property #getter, no setter so read-only
    def redshift_prior_norm(self):
        try:
            # Return the precalculated value
            return self._redshift_prior_norm
        except AttributeError:
            # First time, precalculate and return if max_ref_mag_hi is set
            if self.max_ref_mag_hi is not None:
                self._calculateRedshiftPriorNorm()
                return self._redshift_prior_norm
            else:
                # Complain if it's not
                raise ValueError('Trying to use prior when max_ref_mag_hi is '
                                 + 'not known. If you are using ref_mag_hi_sigma, '
                                 + 'the prior can only be called when set '
                                 + 'as the model inside a Photoz() object. If you '
                                 + 'are using ref_mag_hi, you must not set ref_mag_hi_sigma '
                                 + 'in the config file as it will be preferred.')
    def _loadParameterDict(self):
        nt = len(self.possible_types)

        #if len(self.prior_params) != 5 * len(self.possible_types) - 2:
        if len(self.prior_params) != 5 * len(self.possible_types) - 1:
            raise ValueError('Wrong number of parameters')

        kt = {t: self.prior_params[i] for i, t in enumerate(self.possible_types[:-1])}
        ft = {t: self.prior_params[i + nt - 1] for i, t in enumerate(self.possible_types[:-1])}
        alpt = {t: self.prior_params[i + 2*nt - 2] for i, t in enumerate(self.possible_types)}
        z0t = {t: self.prior_params[i + 3*nt - 2] for i, t in enumerate(self.possible_types)}
        kmt = {t: self.prior_params[i + 4*nt - 2] for i, t in enumerate(self.possible_types)}
        phi = self.prior_params[-1]

        self._prior_params_dict = {
            'k_t': kt, 'f_t': ft, 'alpha_t': alpt, 'z_0t': z0t, 'k_mt': kmt, 'phi':phi
        }

    def _calculateRedshiftPriorNorm(self):
        self._redshift_prior_norm = {}
        mag_range = np.linspace(self.config.ref_mag_lo, self.max_ref_mag_hi, self.mag_grid_len)
        for T in self.possible_types:
            norms = np.zeros(self.mag_grid_len)
            for i, mag in enumerate(mag_range):
                zi = np.exp(np.array([self.lnRedshiftPrior(zz, T, mag, norm=False) for zz in self.responses.zGrid]))
                norms[i] = np.log(1./np.trapz(zi[np.isfinite(zi)], x=self.responses.zGrid[np.isfinite(zi)]))
            self._redshift_prior_norm[T] = interp1d(mag_range, norms)

    def _calculateMagnitudePriorNorm(self, photometry):
        #Integrates over P(m) * S(m) - the selection depends on galaxy
        self.magnitude_prior_norm = np.zeros(photometry.num_galaxies)
        mag_grid = np.linspace(self.config.ref_mag_lo, self.max_ref_mag_hi, self.mag_grid_len)
        flux_grid = 10.**(-0.4 * mag_grid)

        mag_lnPrior = np.array([self.lnMagnitudePrior(m) for m in mag_grid])
        for gal in photometry:
            lnSelection_effect = self.lnSelection(flux_grid, gal)
            prior_select =  np.exp(mag_lnPrior + lnSelection_effect)
            self.magnitude_prior_norm[gal.index] = np.log(np.trapz(prior_select, x=mag_grid))


    def lnTemplatePrior(self, template_type, component_ref_mag):
        mag_diff = component_ref_mag - self.config.ref_mag_lo
        #All include a scaling of 1/Number of templates of that type
        if template_type in self.possible_types[:-1]:
            Nt = self.responses.templates.numType(template_type)
            coeff = np.log(self.prior_params_dict['f_t'][template_type] / Nt)
            expon = self.prior_params_dict['k_t'][template_type] * mag_diff
            out = coeff - expon
        #Prior for final type = 1 - prior of other types
        elif template_type == self.possible_types[-1]:
            Nt = self.responses.templates.numType(template_type)
            other_types = 0.
            for T in self.possible_types[:-1]:
                expon = self.prior_params_dict['k_t'][T] * mag_diff
                other_types += self.prior_params_dict['f_t'][T] * np.exp(-expon)
            out = np.log(1. - other_types) - np.log(Nt)
        else:
            raise ValueError('The possible galaxy types based on your template '
                             'set are "' + '", "'.join(self.possible_types) + '", but the '
                             'template prior was called with type ' + template_type)
        return out

    def lnRedshiftPrior(self, redshift, template_type, component_ref_mag, norm=True):
        try:
            if redshift==0:
                first = -np.inf
            else:
                first = (self.prior_params_dict['alpha_t'][template_type] * np.log(redshift))
            second = self.prior_params_dict['z_0t'][template_type] + (self.prior_params_dict['k_mt'][template_type] * (component_ref_mag - self.config.ref_mag_lo))
            out = first - (redshift / second)**self.prior_params_dict['alpha_t'][template_type]
        except KeyError:
            raise ValueError('The possible galaxy types based on your template '
                             'set are "' + '", "'.join(self.possible_types) + '", but the '
                             'redshift prior was called with type ' + template_type)
        if norm:
            try:
                out = out + self.redshift_prior_norm[template_type](component_ref_mag)
            except ValueError:
                raise ValueError('Magnitude = {} is outside of prior-precalculation '
                                 'range. Check your configuration ref-mag limits '
                                 'cover your input magnitudes.'.format(component_ref_mag))
            return out
        else:
            return out

    def correlationFunction(self, redshifts):
        if len(redshifts)==1:
            return 0.
        elif len(redshifts)==2:
            theta = self.config.angular_resolution
            redshifts = np.sort(redshifts)
            r_2 = self.comovingSeparation(0., redshifts[1])
            delta_r = self.comovingSeparation(redshifts[0], redshifts[1])
            power = 1. - (self.config.gamma/2.)
            one = (self.config.r0**2.) / (power * r_2 * r_2 * theta * theta)
            two = (delta_r**2 + (r_2 * r_2 * theta * theta)) / (self.config.r0**2.)
            three = (delta_r**2) / (self.config.r0**2.)
            return one * ( (two**power) - (three**power) )
        else:
            # Assume any xi^N, N>2 is zero, call two-point function recursively and sum
            return logsumexp([self.correlationFunction(np.asarray(cmb)) for cmb in combinations(redshifts, 2)])

    def lnMagnitudePrior(self, magnitude):
        return (self.prior_params_dict['phi'] * magnitude) * np.log(10.)

    def lnPrior(self, redshift, magnitude):
        #Just the single component prior given TYPE index (not template)
        lnPriorOut = np.zeros(len(self.possible_types))
        for i, template_type in enumerate(self.possible_types):
            p_z = self.lnRedshiftPrior(redshift, template_type, magnitude)
            p_t = self.lnTemplatePrior(template_type, magnitude)
            p_m = self.lnMagnitudePrior(magnitude)
            lnPriorOut[i] = p_z + p_t + p_m
        return lnPriorOut

    def lnPriorCalibrationPrior(self):
        '''Returns the prior on the prior parameters for the calibration procedure.'''
        #Assume a flat prior, except that sum(type fractions) <= 1. ...
        if sum(self.prior_params_dict['f_t'].values()) > 1.:
            return -np.inf
        elif np.any(self.prior_params < 0.):
            return -np.inf
        else:
            return 0.


    def _lnPriorCalibrationPosterior(self, params, photometry):
        self.prior_params = params
        self._loadParameterDict()
        self._calculateRedshiftPriorNorm()
        self._calculateMagnitudePriorNorm(photometry)

        calibration_prior = self.lnPriorCalibrationPrior()

        if not np.isfinite(calibration_prior):
            return -np.inf
        else:
            lnProb_all = 0.
            for g in photometry:
                total_ref_mag = g.ref_mag_data
                magnitude_prior = self.lnMagnitudePrior(total_ref_mag)
                if not np.isfinite(magnitude_prior):
                    pass
                else:
                    total_ref_flux = 10.**(-0.4 * total_ref_mag)
                    selection_effect = self.lnSelection(total_ref_flux, g)
                    template_priors = np.zeros(self.num_templates)
                    redshift_priors = np.zeros(self.num_templates)
                    lnProb_g = -np.inf

                    cache_lnTemplatePrior = {}
                    cache_lnRedshiftPrior = {}
                    for tmpType in self.responses.templates.possible_types:
                        cache_lnTemplatePrior[tmpType] = self.lnTemplatePrior(tmpType, total_ref_mag)
                        cache_lnRedshiftPrior[tmpType] = self.lnRedshiftPrior(g.truth[0]['redshift'], tmpType, total_ref_mag)

                    #Sum over template
                    for T in range(self.num_templates):
                        tmp = 0.

                        tmpType = self.responses.templates.templateType(T)
                        tmp += cache_lnTemplatePrior[tmpType]
                        tmp += cache_lnRedshiftPrior[tmpType]
                        tmp += self._fixed_lnLikelihood_flux[g.index, T]
                        tmp += magnitude_prior
                        tmp += selection_effect
                        #Normalise the magnitude prior
                        tmp -= self.magnitude_prior_norm[g.index]

                        lnProb_g = np.logaddexp(lnProb_g, tmp)
                    lnProb_all += lnProb_g
            if not np.isfinite(lnProb_all):
                return -np.inf
            else:
                return lnProb_all + calibration_prior

    def _negativeLnPriorCalibrationPosterior(self, params, photometry):
        return -1. * self._lnPriorCalibrationPosterior(params, photometry)

    def calibrate(self, photometry, cached_likelihood, mag_grid_len=10,
                  frac_tol=10.,
                  config_save_path='calibrated_prior_config.txt'):

        tolerance = frac_tol * np.finfo(float).eps
        self.mag_grid_len = mag_grid_len
        self._fixed_lnLikelihood_flux = cached_likelihood

        # Set the max_ref_mag_hi (for the precaculation grid) to the
        # largest value in the photometry
        self.max_ref_mag_hi = np.max([g.ref_mag_data for g in photometry])

        # Initial guesses:
        # Assume 1/3 fraction for each type
        # Assume no magnitude dependence --> every k zero
        # With no mag dependence, z0 is the peak of the redshift prior ~ mean(redshifts)
        # 1/alpha controls width of prior ~ 1 / (2 pi std(redshifts))
        redshifts = [g.truth[0]['redshift'] for g in photometry]
        init_z0 = np.mean(redshifts)
        init_a = 1. / (2. * np.pi * np.std(redshifts))
        ntype = len(self.responses.templates.possible_types)

        init_guess = np.array(list(repeat(0., ntype-1)) +
                              list(repeat(1./ntype, ntype-1)) +
                              list(repeat(init_a, ntype)) +
                              list(repeat(init_z0, ntype))+
                              list(repeat(0., ntype)) + [0.6])

        # Bound parameters to be positive, and fractions < 1
        param_bounds = list(repeat((0, None), ntype-1)) + \
                       list(repeat((0, 1), ntype-1)) + \
                       list(repeat((0, None), 3*ntype)) + [(0, 2)]

        results = minimize(self._negativeLnPriorCalibrationPosterior,
                           init_guess, method='L-BFGS-B', jac=False,
                           bounds=param_bounds, tol=tolerance, args=(photometry,),
                           options={'disp':True, 'ftol':tolerance})
        opt_params = results.x

        # Set up self with optimal parameters
        self.prior_params = opt_params
        self._loadParameterDict()
        self._calculateRedshiftPriorNorm()

        #Save out to a config file ready to load into Photoz()
        if config_save_path is not None:
            array_str = np.array2string(opt_params, separator=',', max_line_width=np.inf)[1:-1]
            cfg_str = u'[Run]\n\nprior_params = ' + array_str
            with open(config_save_path, 'w') as config_file:
                config_file.write(cfg_str)



from builtins import *
import warnings
import numpy as np
import blendz
from blendz.config import DefaultConfiguration


'''
Config class description text.
'''

class Configuration(DefaultConfiguration):
    '''
    Container for every setting.
    '''

    def __init__(self, config_path=None, fallback_to_default=True, **kwargs):
        super(Configuration, self).__init__()
        self.default = DefaultConfiguration()
        self.default.populate_configuration()

        self.kwargs = kwargs

        if fallback_to_default:
            self.configs_to_read = self.default.configs_to_read
        else:
            self.configs_to_read = []
        #Add (maybe list of) user configs onto list to read
        if config_path is not None:
            if isinstance(config_path, list):
                self.configs_to_read.extend(config_path)
            else:
                self.configs_to_read.append(config_path)

        self.populate_configuration()

    def keyIsDefault(self, key):
        '''Returns True if the current value of the setting ``key`` is default value'''
        if (key in self.__dict__) and (key in self.default.__dict__) and \
                np.all(self.__dict__[key]==self.default.__dict__[key]):
            return True
        else:
            #Account for hidden properties
            key_mod = '_' + key
            if (key_mod in self.__dict__) and (key_mod in self.default.__dict__) and \
                    np.all(self.__dict__[key_mod]==self.default.__dict__[key_mod]):
                return True
            #Otherwise, not default
            else:
                return False

    def mergeFromOther(self, other_config, overwrite_default_settings=True,
                       overwrite_any_setting=False):
        '''
        Safely merge another configuration object into this one.

        If a setting in the other object is not in this one, it is always merged.
        If it is, whether is is merged or not is controlled by
        `overwrite_default_settings` or `overwrite_nondefault_settings`,
        depending on whether the setting in this object is currently set to
        the default value. If a setting is in both objects but the other is a
        default, a merge is not attempted.
        '''
        no_compare = ['config', 'configs_to_read', 'kwargs', 'default']
        if self.default != other_config.default:
            warnings.warn('Merging configurations with different defaults.')

        for key in other_config.__dict__:
            if key not in no_compare:
                done_set = False
                #Setting not here, merge in
                if not done_set and key not in self.__dict__:
                    self.__dict__[key] = other_config.__dict__[key]
                    done_set = True
                    # ... and mark it as a default if it was a default in other_config
                    if other_config.keyIsDefault(key):
                        self.default.__dict__[key] = other_config.default.__dict__[key]

                #If they're already equal it doesn't matter
                if not done_set and np.all(self.__dict__[key] == other_config.__dict__[key]):
                    done_set = True
                    continue

                #Prevent error if other_config is a default setting and is
                #different to one already here (don't overwrite with defaults)
                if not done_set and other_config.keyIsDefault(key):
                    done_set = True
                    continue

                #Default setting here and allowed to overwrite it
                if not done_set and overwrite_default_settings and self.keyIsDefault(key):
                    self.__dict__[key] = other_config.__dict__[key]
                    done_set = True

                #Allowed to overwrite any setting
                if not done_set and overwrite_any_setting:
                    self.__dict__[key] = other_config.__dict__[key]
                    done_set = True

                #Setting here is nan, which designates it as always okay to overwrite
                if not done_set and self.__dict__[key] is np.nan:
                    self.__dict__[key] = other_config.__dict__[key]
                    done_set = True

                #Setting in other is nan, never overwrite with it
                if not done_set and other_config.__dict__[key] is np.nan:
                    done_set = True

                if not done_set:
                    raise ValueError('Attempting to merge incompatible configurations - ' + key)



from builtins import *
import warnings
from math import ceil
from itertools import repeat
import numpy as np
import emcee
from tqdm import tqdm
from blendz.config import Configuration
from blendz.photometry import PhotometryBase, Galaxy
from blendz.model import BPZ
from blendz.utilities import incrementCount

class SimulatedPhotometry(PhotometryBase):
    def __init__(self, num_sims, config=None, num_components=1, max_redshift=None,
                model=None, seed=None, signal_to_noise=10.,
                num_walkers=100, burn_len=10000,
                measurement_component_specification=None, magnitude_bounds=None, **kwargs):
        super(SimulatedPhotometry, self).__init__(config=config, **kwargs)

        if model is not None:
            #Set model, take default+kwargs+model config and warn user if config also provided
            self.config = Configuration(**kwargs)
            self.config.mergeFromOther(model.config)
            self.model = model
            self.responses = self.model.responses

            if config is not None:
                warnings.warn('A configuration object was provided to '
                              + 'SimulatedPhotometry as well as a Responses '
                              + 'object, though these should be mutually exclusive. '
                              + 'The configuration provided will be ignored.')
        else:
            #Config given, take default+kwargs+given config
            #which is loaded in in PhotometryBase
            self.model = BPZ(config=self.config, max_ref_mag_hi=self.config.ref_mag_hi)
            self.responses = self.model.responses

        #Config must have magnitude_limit set, NOT magnitude_limit_col for simulations
        # We assume a fixed magnitude cut.
        if self.config.magnitude_limit is None:
            raise ValueError('SimulatedPhotometry requires magnitude_limit to be set in config.')
        elif self.config.magnitude_limit_col is not None:
            warnings.warn('SimulatedPhotometry uses magnitude_limit, but magnitude_limit_col has also been set.')

        #Config must have ref_mag_hi set, NOT ref_mag_hi_sigma for simulations
        if self.config.ref_mag_hi is None:
            raise ValueError('SimulatedPhotometry requires ref_mag_hi to be set in config.')
        elif self.config.ref_mag_hi_sigma is not None:
            warnings.warn('SimulatedPhotometry uses ref_mag_hi, but ref_mag_hi_sigma has also been set.')

        self.num_sims = num_sims
        self.num_components = num_components
        self.num_measurements = self.responses.filters.num_filters

        if max_redshift is None:
            self.max_redshift = self.config.z_hi
        else:
            self.max_redshift = max_redshift

        if magnitude_bounds is None:
            self.magnitude_bounds = [self.config.ref_mag_lo, self.config.ref_mag_hi]
        else:
            self.magnitude_bounds = magnitude_bounds

        self.num_walkers = num_walkers
        self.burn_len = burn_len

        self.zero_point_errors = self.config.zero_point_errors
        self.zero_point_frac = 10.**(0.4*self.zero_point_errors) - 1.

        #Signal to noise is an array with an element for each filters
        #If the arg is a float, assume every element the same
        if isinstance(signal_to_noise, float):
            self.signal_to_noise = np.ones(self.model.responses.filters.num_filters) * signal_to_noise
        else:
            self.signal_to_noise = signal_to_noise

        #Use an incrementing counter for the seed to make sure its always
        #different between various different function calls
        if seed is None:
            self.sim_seed = incrementCount(np.random.randint(1e9))
        else:
            self.sim_seed = incrementCount(seed)

        self.simulateRandomGalaxies(self.num_components, self.num_sims,
                                    max_redshift=self.max_redshift,
                                    measurement_component_specification=measurement_component_specification,
                                    magnitude_bounds=self.magnitude_bounds)

        #Remove simulated galaxies that aren't within the selection
        in_selection = np.where(np.array([self._fluxDataWithinSelection(g.flux_data)
                                          for g in self.all_galaxies]))[0]
        self.all_galaxies = [g for i, g in enumerate(self.all_galaxies)
                             if i in in_selection]
        #Go back through and set the galaxy indices to the correct value
        for i, gal in enumerate(self.all_galaxies):
            gal.index = i

    def drawParametersFromPrior(self, num_components, num_sims, burn_len=10000, num_walkers=100, num_thin=50):
        '''
        Use mcmc to draw ``num_sims`` sets of source parameters for
        sources of ``num_components`` components.

        Parameters are returned as an array shape (num_sims, 3*num_components)
        where the second axis is ordered [z1, z2... t1, t2... m0_1, m0_2...].
        Because it's a single array, the template indices are floats, not ints,
        though they have been rounded.
        '''
        #Prior boundary parameters
        z_range = self.config.z_hi - self.config.z_lo
        z_shift = self.config.z_lo
        mag_range = self.config.ref_mag_hi - self.config.ref_mag_lo
        mag_shift = self.config.ref_mag_lo
        tmp_range = self.responses.templates.num_templates - 1.
        tmp_shift = 0.
        #Set up emcee sampler
        num_pars = int(3 * num_components)

        all_ranges = [z for z in repeat(z_range, num_components)] + \
                     [t for t in repeat(tmp_range, num_components)] + \
                     [m for m in repeat(mag_range, num_components)]

        all_shifts = [z for z in repeat(z_shift, num_components)] + \
                     [t for t in repeat(tmp_shift, num_components)] + \
                     [m for m in repeat(mag_shift, num_components)]

        start_pos = [(np.random.random(num_pars) * np.array(all_ranges)) + \
                    np.array(all_shifts) for i in range(num_walkers)]

        for w in range(num_walkers):
            if self.config.sort_redshifts:
                #Force sorted redshifts
                start_pos[w][:num_components] = np.sort(start_pos[w][:num_components])
            else:
                #Force sorted magnitudes
                start_pos[w][-num_components:] = np.sort(start_pos[w][-num_components:])

        sampler = emcee.EnsembleSampler(num_walkers, num_pars, self.model._lnTotalPrior)
        #Run burn in
        burn_sample_len = int(ceil(burn_len / float(num_walkers)))
        burn_rstate = np.random.RandomState(next(self.sim_seed))
        burn_pos, burn_prob, burn_state = sampler.run_mcmc(start_pos,
                                                           burn_sample_len,
                                                           rstate0=burn_rstate)
        #Sample our parameters until we have enough selected sources
        num_selected = 0
        #Sample in batches of num_sims samples, plus a bit more
        #because some are lost due to not making the selection criteria
        main_sample_len = int(ceil(num_sims / float(num_walkers)) * 1.2 * num_thin)
        params = np.zeros((num_sims, num_pars))
        #Fill up the return params array with samples from the chain
        #that obey the magnitude selection criteria
        with tqdm(total=num_sims) as pbar:
            while num_selected < num_sims:
                sampler.reset()
                main_rstate = np.random.RandomState(next(self.sim_seed))
                sampler.run_mcmc(burn_pos, main_sample_len,
                                 thin=num_thin, rstate0=main_rstate)
                chain = sampler.flatchain
                for i in range(np.shape(chain)[0]):
                    sample_i = chain[-i, :]
                    if num_selected<num_sims:# and self._fluxDataWithinSelection(sample_i):
                        params[num_selected, :] = sample_i
                        params[num_selected, num_components:2*num_components] = \
                            np.around(params[num_selected, num_components:2*num_components])
                        num_selected += 1
                        pbar.update()


        return params

    def _fluxDataWithinSelection(self, flux_data):
        '''Take in an array of flux data and return a bool of whether
        is is within the selection criteria
        '''
        ref_band_mag = np.log10(flux_data[self.config.ref_band]) / (-0.4)
        return ref_band_mag <= self.config.magnitude_limit

    def generateObservables(self, params):
        '''
        Use array of params shape (num_sims, 3*num_components) to generate
        array of observed fluxes and array of errors, both of shape
        (num_sims, num_filters)
        '''
        num_sims, num_params = np.shape(params)
        num_components = num_params // 3
        out_shape = (num_sims, self.responses.filters.num_filters)
        true_flux = np.zeros(out_shape)
        for g in range(num_sims):
            for c in range(num_components):
                zc = params[g, c]
                tc = int(params[g, num_components + c])
                mc = params[g, (2*num_components) + c]
                resp_c = self.responses.interp(zc)[tc, :]
                norm = (10.**(-0.4 * mc)) / resp_c[self.config.ref_band]
                true_flux[g, :] += resp_c * norm * self.model.measurement_component_mapping[c, :]

        #Errors
        flux_err = true_flux / self.signal_to_noise #quoted error
        err_rstate = np.random.RandomState(next(self.sim_seed))
        rand_err = err_rstate.normal(loc=0., scale=flux_err, size=out_shape) #actual error
        obs_flux = true_flux + rand_err #added actual error to flux

        #Conversions
        obs_mag = np.log10(obs_flux) / (-0.4)
        mag_err = np.log10((flux_err/obs_flux)+1.) / 0.4
        return obs_mag, mag_err

    def createTruthDicts(self, params):
        '''
        Use array of params shape (num_sims, 3*num_components) to generate
        list of truth dictionaries.
        '''
        num_sims, num_params = np.shape(params)
        num_components = num_params // 3

        all_truths = []
        for g in range(num_sims):
            redshifts = params[g, :num_components]
            templates = params[g, num_components:2*num_components].astype(int)
            magnitudes = params[g, 2*num_components:]
            #Order component truths before saving, either by redshift...
            if self.config.sort_redshifts:
                order = np.argsort(redshifts)
                redshifts = redshifts[order]
                templates = templates[order]
                magnitudes = magnitudes[order]
            #...or magnitudes
            else:
                order = np.argsort(magnitudes)
                redshifts = redshifts[order]
                templates = templates[order]
                magnitudes = magnitudes[order]

            truth = {}
            truth['num_components'] = num_components
            for c in range(num_components):
                truth[c] = {'redshift': redshifts[c],
                            'template': templates[c],
                            'magnitude': magnitudes[c]}
            all_truths.append(truth)
        return all_truths

    def simulateRandomGalaxies(self, num_components, num_sims, max_redshift=None,
                               measurement_component_specification=None,
                               magnitude_bounds=None, burn_len=None, num_walkers=None):
        if max_redshift is None:
            max_redshift = self.max_redshift
        if magnitude_bounds is None:
            magnitude_bounds = self.magnitude_bounds
        if burn_len is None:
            burn_len = self.burn_len
        if num_walkers is None:
            num_walkers = self.num_walkers

        self.model._setMeasurementComponentMapping(measurement_component_specification, num_components)

        prior_parameters = self.drawParametersFromPrior(num_components, num_sims, burn_len=burn_len, num_walkers = num_walkers)
        all_mag_data, all_mag_sigma = self.generateObservables(prior_parameters)
        all_truths = self.createTruthDicts(prior_parameters)

        for g in range(num_sims):
            new_galaxy = Galaxy(all_mag_data[g, :], all_mag_sigma[g, :], self.config, self.zero_point_frac, g)
            new_galaxy.truth = all_truths[g]
            new_galaxy.magnitude_limit = self.config.magnitude_limit
            new_galaxy.ref_mag_hi = self.config.ref_mag_hi
            self.all_galaxies.append(new_galaxy)



try:
    from mpi4py import MPI
    MPI_RANK = MPI.COMM_WORLD.Get_rank()
except:
    MPI_RANK = 0

from builtins import *
import sys
import os
import warnings
from math import ceil
from multiprocessing import cpu_count
import itertools as itr
import numpy as np
from scipy.integrate import simps
import nestle
import emcee
from tqdm import tqdm
import dill
import blendz
from blendz import Configuration
from blendz.fluxes import Responses
from blendz.photometry import Photometry, SimulatedPhotometry
from blendz.utilities import incrementCount, Silence

try:
    import pymultinest
    PYMULTINEST_AVAILABLE = True
except ImportError:
    PYMULTINEST_AVAILABLE = False
    warnings.warn('PyMultinest not installed, so falling back to (slower) python implementation.'
        + ' See http://johannesbuchner.github.com/PyMultiNest/install.html for installation help.')
except (SystemExit, OSError):
    PYMULTINEST_AVAILABLE = False
    warnings.warn('PyMultinest failed to load, so falling back to (slower) python implementation.'
        + ' See http://johannesbuchner.github.com/PyMultiNest/install.html for installation help.')


class Photoz(object):
    def __init__(self, model=None, photometry=None, config=None,\
                 load_state_path=None, **kwargs):
        if load_state_path is not None:
            self.loadState(load_state_path)
        else:
            #Warn user is config and either/or responses/photometry given that config ignored
            if ((model is not None and config is not None) or
                    (photometry is not None and config is not None)):
                warnings.warn('A configuration object was provided to Photoz object '
                              + 'as well as a Model/Photometry object, though these '
                              + 'should be mutually exclusive. The configuration '
                              + 'provided will be ignored.')
            #Responses and photometry given, merge their configs
            if (model is not None) and (photometry is not None):
                self.config = Configuration(**kwargs)
                self.config.mergeFromOther(model.config)
                self.config.mergeFromOther(photometry.config)
                self.model = model
                self.responses = self.model.responses
                self.photometry = photometry
            #Only responses given, use its config to load photometry
            elif (model is not None) and (photometry is None):
                self.config = Configuration(**kwargs)
                self.config.mergeFromOther(model.config)
                self.model = model
                self.responses = self.model.responses
                self.photometry = Photometry(config=self.config)
            #Only photometry given, use its config to load responses
            elif (model is None) and (photometry is not None):
                self.config = Configuration(**kwargs)
                self.config.mergeFromOther(photometry.config)
                self.photometry = photometry

                self.model = blendz.model.BPZ(config=self.config)
                self.responses = self.model.responses
            #Neither given, load both from provided (or default, if None) config
            else:
                self.config = Configuration(**kwargs)
                if config is not None:
                    self.config.mergeFromOther(config)

                self.photometry = Photometry(config=self.config)

                self.model = blendz.model.BPZ(config=self.config)
                self.responses = self.model.responses

            # Get max_ref_mag_hi from photometry, which already deals with sigma vs fixed
            self.max_ref_mag_hi = np.max([g.ref_mag_hi for g in self.photometry])
            self.model.max_ref_mag_hi = self.max_ref_mag_hi

            self.num_templates = self.responses.templates.num_templates
            self.num_measurements = self.responses.filters.num_filters
            self.num_galaxies = self.photometry.num_galaxies

            self.tmp_ind_to_type_ind = self.responses.templates.tmp_ind_to_type_ind
            self.possible_types = self.responses.templates.possible_types
            self.num_types = len(self.possible_types)

            #Default to assuming single component, present in all measurements
            self.model._setMeasurementComponentMapping(None, 1)

            #Set up empty dictionaries to put results into
            self._samples = {}
            self._logevd = {}
            self._logevd_error = {}
            for g in range(self.num_galaxies):
                #Each value is a dictionary which will be filled by sample function
                #The keys of this inner dictionary will be the number of blends for run
                self._samples[g] = {}
                self._logevd[g] = {}
                self._logevd_error[g] = {}

    def saveState(self, filepath):
        """Save this entire Photoz instance to file.

        This saves the exact state of the current object, including all data and any
        reults from sampling.

        Args:
            filepath (str): Path to file to save to.
        """
        if isinstance(self.photometry, SimulatedPhotometry):
            try:
                current_seed = self.photometry.sim_seed.next()
                self.photometry.sim_seed = current_seed
            except:
                warnings.warn('SimulatedPhotometry seed not saved.')
        with open(filepath, 'wb') as f:
            state = {key: val for key, val in self.__dict__.items() if key!='pbar'}
            dill.dump(state, f)
        #Put the random seed back how it was after the saving is done
        if isinstance(self.photometry, SimulatedPhotometry):
            try:
                self.photometry.sim_seed = incrementCount(current_seed)
            except:
                pass

    def loadState(self, filepath):
        with open(filepath, 'rb') as f:
            self.__dict__.update(dill.load(f))
        #If the photometry is simulated, replace the seed currently saved as
        #a number with the generator it was before saving
        if isinstance(self.photometry, SimulatedPhotometry):
            try:
                current_seed = self.photometry.sim_seed
                self.photometry.sim_seed = incrementCount(current_seed)
            except:
                warnings.warn('SimulatedPhotometry seed not loaded.')

    def _lnLikelihood_flux(self, model_flux):
        chi_sq = -1. * np.sum((self.photometry.current_galaxy.flux_data_noRef - model_flux)**2 / self.photometry.current_galaxy.flux_sigma_noRef**2)
        return chi_sq

    def _lnLikelihood_mag(self, total_ref_flux):
        #chi_sq = -1. * np.sum((self.photometry.current_galaxy.ref_mag_data - total_ref_mag)**2 / self.photometry.current_galaxy.ref_mag_sigma**2)
        chi_sq = -1. * np.sum((self.photometry.current_galaxy.ref_flux_data - total_ref_flux)**2 / self.photometry.current_galaxy.ref_flux_sigma**2)
        return chi_sq

    def _lnPosterior(self, params):
        num_components = int(len(params) // 2)
        redshifts = params[:num_components]
        magnitudes = params[num_components:]

        if not self.model._obeyPriorConditions(redshifts, magnitudes, self.photometry.current_galaxy.ref_mag_hi):
            return -np.inf
        else:
            #Precalculate all quantities we'll need in the template loop
            #Single interp call -> Shape = (N_template, N_band, N_component)
            model_fluxes = self.responses.interp(redshifts)

            priors = np.zeros((num_components, self.num_types))
            for nb in range(num_components):
                priors[nb, :] = self.model.lnPrior(redshifts[nb], magnitudes[nb])

            redshift_correlation = np.log(1. + self.model.correlationFunction(redshifts))

            #Get total flux in reference band  = transform to flux & sum
            total_ref_flux = np.sum(10.**(-0.4 * magnitudes))
            selection_effect = self.model.lnSelection(total_ref_flux,
                                                      self.photometry.current_galaxy)

            #Loop over all templates - discrete marginalisation
            #All log probabilities so (multiply -> add) and (add -> logaddexp)
            lnProb = -np.inf

            #At each iteration template_combo is a tuple of (T_1, T_2... T_num_components)
            for template_combo in itr.product(*itr.repeat(range(self.num_templates), num_components)):
                #One redshift/template/magnitude prior and model flux for each blend component
                tmp = 0.
                blend_flux = np.zeros(self.num_measurements)
                component_scaling_norm = 0.
                for nb in range(num_components):
                    T = template_combo[nb]
                    component_scaling = 10.**(-0.4*magnitudes[nb]) / model_fluxes[T, self.config.ref_band, nb]
                    blend_flux += model_fluxes[T, :, nb] * component_scaling * self.model.measurement_component_mapping[nb, :]
                    type_ind = self.tmp_ind_to_type_ind[T]
                    tmp += priors[nb, type_ind]
                #Remove ref_band from blend_fluxes, as that goes into the magnitude
                #likelihood, not the flux likelihood
                blend_flux = blend_flux[self.config.non_ref_bands]

                #Other terms only appear once per summation-step
                tmp += redshift_correlation
                tmp += self._lnLikelihood_flux(blend_flux)
                tmp += self._lnLikelihood_mag(total_ref_flux)
                tmp += selection_effect

                #logaddexp contribution from this template to marginalise
                lnProb = np.logaddexp(lnProb, tmp)

            return lnProb - self.prior_norm


    def _priorTransform(self, params):
        '''
        Transform params from [0, 1] uniform random to [min, max] uniform random,
        where the min redshift is zero, max redshift is set in the configuration,
        and the min/max magnitudes (numerically, not brightness) are set by configuration.
        '''
        num_components = int(len(params) // 2)

        trans = np.ones(len(params))
        trans[:num_components] = self.config.z_hi - self.config.z_lo
        trans[num_components:] = self.photometry.current_galaxy.ref_mag_hi - self.config.ref_mag_lo

        shift = np.zeros(len(params))
        shift[:num_components] = self.config.z_lo
        shift[num_components:] = self.config.ref_mag_lo
        return (params * trans) + shift


    def _priorTransform_multinest(self, cube, ndim, nparams):
        '''
        Transform params from [0, 1] uniform random to [0, max] uniform random,
        where the max redshift is set in the configuration, and the max fraction
        is 1.
        '''
        num_components = ndim // 2

        trans = np.zeros(ndim)
        trans[:num_components] = self.config.z_hi
        trans[num_components:] = self.photometry.current_galaxy.ref_mag_hi - self.config.ref_mag_lo

        shift = np.zeros(ndim)
        shift[:num_components] = self.config.z_lo
        shift[num_components:] = self.config.ref_mag_lo

        for i in range(ndim):
            cube[i] = (cube[i] * trans[i]) + shift[i]

    def _lnPosterior_multinest(self, cube, ndim, nparams):
        self.num_posterior_evals += 1

        params = np.array([cube[i] for i in range(ndim)])

        with self.breakSilence():
            if (self.num_posterior_evals%self.num_between_print==0) and MPI_RANK==0:
                self.pbar.set_description('[Cmp: {}/{}, Itr: {}] '.format(self.blend_count,
                                                                          self.num_components_sampling,
                                                                          self.num_posterior_evals))
                self.pbar.refresh()

        return self._lnPosterior(params)

    def _sampleProgressUpdate(self, info):
        if (info['it']%self.num_between_print==0) and MPI_RANK==0:
            self.pbar.set_description('[Cmp: {}/{}, Itr: {}] '.format(self.blend_count,
                                                                      self.num_components_sampling,
                                                                      info['it']))
            self.pbar.refresh()

    def normalise_prior_setup(self, galind, num_components, magnitude_grid_len=25,
                              redshift_grid_len=10):
        # Set up 1D grids
        self.magnitude_grid_len = magnitude_grid_len
        self.redshift_grid_len = redshift_grid_len
        self.redshift_grid = np.linspace(self.config.z_lo, self.config.z_hi,
                                         self.redshift_grid_len)
        ref_mag_hi = self.photometry[galind].ref_mag_hi
        self.magnitude_grid = np.linspace(self.config.ref_mag_lo, ref_mag_hi,
                                          self.magnitude_grid_len)

        # Create 3D grid for single-component prior - redshift x magnitude x template
        prior_grid_type = np.zeros((self.redshift_grid_len, self.magnitude_grid_len, self.num_types))
        for i, z in enumerate(self.redshift_grid):
            for j, m in enumerate(self.magnitude_grid):
                prior_grid_type[i, j, :] = np.exp(self.model.lnPrior(z, m))

        self.prior_grid_tmp = np.zeros((self.redshift_grid_len, self.magnitude_grid_len, self.num_templates))
        for tmp in range(self.num_templates):
            t = self.tmp_ind_to_type_ind[tmp]
            self.prior_grid_tmp[:, :, tmp] = prior_grid_type[:, :, t]

    def normalise_prior(self, galind, num_components):
        # This could/should be replaced with code to cache prior grid
        self.normalise_prior_setup(galind, num_components)

        # Create N-D grid of 1 + xi([za, zb...])
        xi_grid = np.zeros(tuple(self.redshift_grid_len for
                            i in range(num_components)))
        for inds in itr.product(*itr.repeat(range(self.redshift_grid_len), num_components)):
            try:
                xi_grid[inds] = 1. + self.model.correlationFunction(
                                    self.redshift_grid[np.array(inds)])
            except ZeroDivisionError:
                xi_grid[inds] = 1.

        # Create N-D grid of S([ma, mb...])
        select_grid = np.zeros(tuple(self.magnitude_grid_len for
                            i in range(num_components)))
        for inds in itr.product(*itr.repeat(range(self.magnitude_grid_len), num_components)):
            total_ref_flux = np.sum(10.**(-0.4*self.magnitude_grid[np.array(inds)]))
            select_grid[inds] = np.exp(self.model.lnSelection(total_ref_flux,
                                                       self.photometry[galind]))
        # Combine N 2D prior grids, N-D xi grid and N-D select grid into one 2N-D grid
        # Axes = za, zb, ... ma, mb ...
        total_shape = tuple(element for tpl in
                            tuple((self.redshift_grid_len, self.magnitude_grid_len) for i in range(num_components))
                            for element in tpl)
        self.total_prior_grid = np.zeros(total_shape)

        for zinds in itr.product(*itr.repeat(range(self.redshift_grid_len), num_components)):
            for minds in itr.product(*itr.repeat(range(self.magnitude_grid_len), num_components)):

                okayz = (self.config.sort_redshifts and tuple(sorted(zinds))==zinds)
                okaym = (not self.config.sort_redshifts) and tuple(sorted(minds))==minds
                if okayz or okaym:
                    prior_prod = np.ones(self.num_templates)
                    for cmp in range(num_components):
                        prior_prod *= self.prior_grid_tmp[zinds[cmp], minds[cmp], :]
                    prior_prod = np.sum(prior_prod)

                    inds = tuple(element for tpl in
                                    tuple((zinds[i], minds[i]) for i in range(num_components))
                                    for element in tpl)

                    self.total_prior_grid[inds] = prior_prod * select_grid[minds] * xi_grid[zinds]
                else:
                    self.total_prior_grid[inds] = 0.

        self.prior_norm = simps(simps(self.total_prior_grid, x=self.magnitude_grid), x=self.redshift_grid)
        for cmp in range(num_components-1):
            self.prior_norm  = simps(simps(self.prior_norm , x=self.magnitude_grid), x=self.redshift_grid)
        self.prior_norm = np.log(self.prior_norm)


    def sample(self, num_components, galaxy=None, nresample=1000, seed=False,
               measurement_component_mapping=None, npoints=150, print_interval=10,
               use_pymultinest=None, save_path=None, save_interval=None):
        """Sample the posterior for a particular number of components.

        Args:
            num_components (int):
                Sample the posterior defined for this number of components in the source.

            galaxy (int or None):
                Index of the galaxy to sample. If None, sample every galaxy in the
                photometry. Defaults to None.

            nresample (int):
                Number of non-weighted samples to draw from the weighted samples
                distribution from Nested Sampling. Defaults to 1000.

            seed (bool or int):
                Random seed for sampling to ensure deterministic results when
                ampling again. If False, do not seed. If True, seed with value
                derived from galaxy index. If int, seed with specific value.

            measurement_component_mapping (None or list of tuples):
                If None, sample from the fully blended posterior. For a partially
                blended posterior, this should be a list of tuples (length = number of
                measurements), where each tuples contains the (zero-based) indices of
                the components that measurement contains. Defaults to None.

            npoints (int):
                Number of live points for the Nested Sampling algorithm. Defaults to 150.

            print_interval (int):
                Update the progress bar with number of posterior evaluations every
                print_interval calls. Defaults to 10.

            save_path (None or str):
                Filepath for saving the Photoz object for reloading with `Photoz.loadState`.
                If None, do not automatically save. If given, the Photoz object will
                be saved to this path after all galaxies are sampled. If save_interval
                is also not None, the Photoz object will be saved to this path every
                save_interval galaxies. Defaults to None.

            save_interval (None or int)
                If given and save_path is not None, the Photoz object will be
                saved to save_path every save_interval galaxies. Defaults to None.

            use_pymultinest (bool or None)
                If True, sample using the pyMultinest sampler. This requires PyMultiNest
                to be installed separately. If False, sample using the Nestle sampler,
                which is always installed when blendz is. If None, check whether pyMultinest
                is installed and use it if it is, otherwise use Nestle. Defaults to None.
        """

        if use_pymultinest is None:
            use_pymultinest = PYMULTINEST_AVAILABLE

        if isinstance(num_components, int):
            num_components = [num_components]
        else:
            if measurement_component_mapping is not None:
                #TODO: This is a time-saving hack to avoid dealing with multiple specifications
                #The solution would probably be to rethink the overall design
                raise ValueError('measurement_component_mapping cannot be set when sampling multiple numbers of components in one call. Do the separate cases separately.')

        self.num_components_sampling = len(num_components)
        self.num_between_print = float(round(print_interval))

        if galaxy is None:
            start = None
            stop = None
            self.num_galaxies_sampling = self.num_galaxies
        elif isinstance(galaxy, int):
            start = galaxy
            stop = galaxy + 1
            self.num_galaxies_sampling = 1
        else:
            raise TypeError('galaxy may be either None or an integer, but got {} instead'.format(type(galaxy)))

        with tqdm(total=self.num_galaxies_sampling, unit='galaxy') as self.pbar:
            self.gal_count = 1
            for gal in self.photometry.iterate(start, stop):
                self.blend_count = 1
                for nb in num_components:

                    if seed is False:
                        rstate = np.random.RandomState()
                    elif seed is True:
                        rstate = np.random.RandomState(gal.index)
                    else:
                        rstate = np.random.RandomState(seed + gal.index)

                    num_param = 2 * nb
                    self.model._setMeasurementComponentMapping(measurement_component_mapping, nb)

                    self.normalise_prior(gal.index, nb)

                    if use_pymultinest:
                        if not os.path.exists('chains'):
                            os.makedirs('chains')
                        with Silence() as self.breakSilence:
                            self.num_posterior_evals = 0
                            pymultinest.run(self._lnPosterior_multinest, self._priorTransform_multinest,
                                            num_param, resume=False, verbose=False, sampling_efficiency='model',
                                            n_live_points=npoints)#,
                                            #outputfiles_basename=os.path.join(blendz.CHAIN_PATH, 'chain_'))
                            results = pymultinest.analyse.Analyzer(num_param)#, outputfiles_basename=os.path.join(blendz.CHAIN_PATH, 'chain_'))

                        self._samples[gal.index][nb] = results.get_equal_weighted_posterior()[:, :-1]
                        self._logevd[gal.index][nb] = results.get_mode_stats()['global evidence']
                        self._logevd_error[gal.index][nb] = results.get_mode_stats()['global evidence error']

                    else:
                        results = nestle.sample(self._lnPosterior, self._priorTransform,
                                                num_param, method='multi', npoints=npoints,
                                                rstate=rstate, callback=self._sampleProgressUpdate)
                        self._samples[gal.index][nb] = results.samples[rstate.choice(len(results.weights), size=nresample, p=results.weights)]
                        self._logevd[gal.index][nb] = results.logz
                        self._logevd_error[gal.index][nb] = results.logzerr

                    self.blend_count += 1

                self.gal_count += 1
                if MPI_RANK==0:
                    self.pbar.update()
                if (save_path is not None) and (save_interval is not None):
                    if gal.index % save_interval == 0:
                        self.saveState(save_path)
        if save_path is not None:
            self.saveState(save_path)

    def _cacheTruthLikelihood(self):
        self.model._setMeasurementComponentMapping(None, 1)
        #Single interp call -> Shape = (N_template, N_band, N_component)
        fixed_model_fluxes = {}
        fixed_lnLikelihood_flux = np.zeros((self.photometry.num_galaxies,
                                            self.responses.templates.num_templates))
        for g in self.photometry:
            fixed_model_fluxes[g.index] = self.responses.interp(np.array([g.truth[0]['redshift']]))
            for T in range(self.responses.templates.num_templates):
                scaling = 10.**(-0.4*g.ref_mag_data) / fixed_model_fluxes[g.index][T, self.config.ref_band, 0]
                fixed_model_fluxes[g.index][T, :, 0] *= scaling
                #Cache the flux likelihoods
                non_ref_flux = fixed_model_fluxes[g.index][T, self.config.non_ref_bands, 0]
                fixed_lnLikelihood_flux[g.index, T] = self._lnLikelihood_flux(non_ref_flux)
        return fixed_lnLikelihood_flux


    def calibrate(self, **kwargs):
        cached_likelihood = self._cacheTruthLikelihood()
        self.model.calibrate(self.photometry, cached_likelihood, **kwargs)

    def chain(self, num_components, galaxy=None):
        """Return the (unweighted) posterior samples.

        Args:
            num_components (int):
                Number of components.

            galaxy (int or None):
                Index of the galaxy to calculate log-evidence for. If None, return array
                of log-evidence for every galaxy. Defaults to None.
        """
        if galaxy is None:
            return [self._samples[g][num_components] for g in range(self.num_galaxies)]
        else:
            return self._samples[galaxy][num_components]

    def logevd(self, num_components, galaxy=None, return_error=False):
        """Return the natural log of the evidence.

        Args:
            num_components (int):
                Number of components.

            galaxy (int or None):
                Index of the galaxy to calculate log-evidence for. If None, return array
                of log-evidence for every galaxy. Defaults to None.

            return_error (bool):
                If True, also return the error on the log-evidence. If galaxy is None, this
                is also an array. Defaults to False.
        """
        if galaxy is None:
            if return_error:
                return np.array([self._logevd[g][num_components] for g in range(self.num_galaxies)]),\
                            np.array([self._logevd_error[g][num_components] for g in range(self.num_galaxies)])
            else:
                return np.array([self._logevd[g][num_components] for g in range(self.num_galaxies)])
        else:
            if return_error:
                return self._logevd[galaxy][num_components], self._logevd_error[galaxy][num_components]
            else:
                return self._logevd[galaxy][num_components]

    def logbayes(self, m, n, base=None, galaxy=None):
        """Return the log of the Bayes factor between m and n components, log[B_mn].

        A positive value suggests that that evidence prefers the m-component model over
        the n-component model.

        Args:
            m (int):
                First number of components.

            n (int):
                Second number of components.

            base (float or None):
                Base of the log to return. If None, uses natural log.
                Defaults to None.

            galaxy (int or None):
                Index of the galaxy to calculate B_mn for. If None, return array
                of B_mn for every galaxy. Defaults to None.
        """
        if base is None:
            alter_base = 1.
        else:
            alter_base = np.log(base)
        return (self.logevd(m, galaxy=galaxy) - self.logevd(n, galaxy=galaxy)) / alter_base

    def applyToMarginals(self, func, num_components, galaxy=None, **kwargs):
        """Apply a function to the 1D marginal distribution samples of each parameter.

        Args:
            func (function):
                The function to apply to the marginal distribution samples.
                It should accept an array of the samples as its first argument,
                with optional keyword arguments.

            num_components (int):
                Number of components.

            galaxy (int or None):
                Index of the galaxy to apply the function to. If None, return
                array with a row for each galaxy. Defaults to None.

            **kwargs:
                Any optional keyword arguments to pass to the function.
        """
        if galaxy is None:
            out = np.zeros((self.num_galaxies, num_components * 2))
            for g in range(self.num_galaxies):
                for n in range(num_components * 2):
                    out[g, n] = func(self._samples[g][num_components][:, n], **kwargs)
        else:
            out = np.zeros(num_components * 2)
            for n in range(num_components * 2):
                out[n] = func(self._samples[galaxy][num_components][:, n], **kwargs)
        return out

    def _MAP1d(self, samps, bins=50):
        vals, edges = np.histogram(samps, bins=bins)
        return edges[np.argmax(vals)] + ((edges[1] - edges[0]) / 2.)

    def max(self, num_components, galaxy=None, bins=20):
        """Return the maximum-a-posteriori point for the 1D marginal distribution of each parameter.

        This is calculated by forming a 1D histogram of each marginal distribution
        and assigning the MAP of that parameter as the centre of the tallest bin.

        Args:
            num_components (int):
                Number of components.

            galaxy (int or None):
                Index of the galaxy to calculate the MAP for. If None, return array
                with rows of MAPs for each galaxy. Defaults to None.

            bins (int):
                Number of bins to use for each 1D histogram.
        """
        return self.applyToMarginals(self._MAP1d, num_components, galaxy=galaxy, bins=bins)

    def mean(self, num_components, galaxy=None):
        """Return the mean point for the 1D marginal distribution of each parameter.

        Args:
            num_components (int):
                Number of components.

            galaxy (int or None):
                Index of the galaxy to calculate the MAP for. If None, return array
                with rows of means for each galaxy. Defaults to None.
        """
        return self.applyToMarginals(np.mean, num_components, galaxy=galaxy)


    def std(self, num_components, galaxy=None):
        """Return the standard deviation for the 1D marginal distribution of each parameter.

        Args:
            num_components (int):
                Number of components.

            galaxy (int or None):
                Index of the galaxy to calculate the MAP for. If None, return array
                with rows of means for each galaxy. Defaults to None.
        """
        return self.applyToMarginals(np.std, num_components, galaxy=galaxy)

    def quantiles(self, num_components, galaxy=None, q=(0.16, 0.84)):
        try:
            pcnt = [qq*100. for qq in q]
        except TypeError:
            pcnt = [100. * q]
        #call numpy.percentile with q *= 100
        return [self.applyToMarginals(np.percentile, num_components, galaxy=galaxy, q=pp) for pp in pcnt]



from builtins import *
import warnings
import numpy as np
from tqdm import tqdm
from scipy.interpolate import interp1d
from blendz import Configuration
from blendz.fluxes import Templates
from blendz.fluxes import Filters

class Responses(object):
    def __init__(self, templates=None, filters=None, config=None, **kwargs):
        #Warn user is config and either/or templates given that config ignored
        if ((templates is not None and config is not None) or
                (filters is not None and config is not None)):
            warnings.warn('A configuration object was provided to Responses '
                          + 'as well as a Template/Filter object, though these '
                          + 'should be mutually exclusive. The configuration '
                          + 'provided will be ignored.')
        #Both templates and filters given, merge with default+kwargs
        if (templates is not None) and (filters is not None):
            self.config = Configuration(**kwargs)
            self.config.mergeFromOther(templates.config)
            self.config.mergeFromOther(filters.config)
            self.templates = templates
            self.filters = filters
        #Templates given but filters not, load filters using default+kwargs+templates config
        elif (templates is not None) and (filters is None):
            self.config = Configuration(**kwargs)
            self.config.mergeFromOther(templates.config)
            self.templates = templates
            self.filters = Filters(config=self.config)
        #Filters given but templates not, load templates using default+kwargs+filters config
        elif (templates is None) and (filters is not None):
            self.config = Configuration(**kwargs)
            self.config.mergeFromOther(filters.config)
            self.filters = filters
            self.templates = Templates(config=self.config)
        #Neither given, load both from provided (or default, if None) config
        else:
            self.config = Configuration(**kwargs)
            if config is not None:
                self.config.mergeFromOther(config)

            self.templates = Templates(config=self.config)
            self.filters = Filters(config=self.config)

        self.zGrid = self.config.redshift_grid
        self._calculate_responses()
        self._calculate_interpolators()

    def _calculate_interpolators(self):
        self._interpolators = {}
        for T in range(self.templates.num_templates):
            self._interpolators[T] = {}
            for F in range(self.filters.num_filters):
                self._interpolators[T][F] = interp1d(self.zGrid, self._all_responses[T, F, :],\
                                                     bounds_error=False, fill_value=0.)

    def _calculate_responses(self):
        self._all_responses = np.zeros((self.templates.num_templates, self.filters.num_filters, len(self.zGrid)))
        tot_its = self.templates.num_templates * self.config.z_len * \
                    self.filters.num_filters
        with tqdm(total=tot_its) as pbar:
            for F in range(self.filters.num_filters):
                for iZ, Z in enumerate(self.zGrid):
                    for T in range(self.templates.num_templates):
                        shiftedTemplate = self.templates.interp(T, self.filters.wavelength(F) / (1+Z) )
                        flux_norm = self.filters.norm(F) * 2.99792458e18
                        integrand = shiftedTemplate * self.filters.response(F) * \
                                    self.filters.wavelength(F) / flux_norm
                        self._all_responses[T, F, iZ] = np.trapz(integrand, x=self.filters.wavelength(F))
                        pbar.update()
        self.interp = interp1d(self.zGrid, self._all_responses, bounds_error=False, fill_value=0.)



from setuptools import setup, find_packages

setup(name = 'blendz',
      version = '1.0.0',
      description = 'Bayesian photometric redshifts of blended sources.',
      author = 'Daniel Michael Jones',
      author_email = 'd.jones15@imperial.ac.uk',
      license = 'MIT',
      url = 'http://blendz.readthedocs.io',
      packages = find_packages(),
      install_requires=[
        'numpy',
        'scipy',
        'tqdm',
        'nestle',
        'dill',
        'future',
        'emcee',
      ],
      include_package_data = True)#,
#      zip_safe = False)



from builtins import *
#Python 2 & 3 compatibility for abstract base classes, from
#https://stackoverflow.com/questions/35673474/
import sys
import warnings
import abc
from future.utils import with_metaclass
import numpy as np
from scipy.integrate import quad
from scipy.special import erf
from blendz import Configuration
from blendz.fluxes import Responses

class ModelBase(with_metaclass(abc.ABCMeta)):
    def __init__(self, responses=None, config=None, **kwargs):
        #Warn user is config and responses given that config ignored
        if ((responses is not None) and (config is not None)):
            warnings.warn('A configuration was provided to Model object '
                          + 'in addition to the Responses, though these '
                          + 'should be mutually exclusive. The configuration '
                          + 'provided will be ignored.')
        #If responses given, use that
        if responses is not None:
            self.config = Configuration(**kwargs)
            self.config.mergeFromOther(responses.config)
            self.responses = responses
        #Otherwise use config to create a responses object
        else:
            self.config = Configuration(**kwargs)
            if config is not None:
                self.config.mergeFromOther(config)

            self.responses = Responses(config=self.config)

        self.prior_params = self.config.prior_params
        self.num_templates = self.responses.templates.num_templates
        self.possible_types = self.responses.templates.possible_types

        #If ref_mag_hi_sigma is not set, use ref_mag_hi instead, but prefer sigma
        if self.config.ref_mag_hi_sigma is None:
            if self.config.ref_mag_hi is not None:
                # Use fixed ref_mag_hi
                self.max_ref_mag_hi = self.config.ref_mag_hi
            else:
                # None available, so raise exception
                raise ValueError('One of ref_mag_hi or ref_mag_hi_sigma must be '
                                 + 'set in config to use the model.')
        else:
            # Using ref_mag_hi_sigma needs photometry, so set to None and
            # let Photoz() deal with setting it once model+photometry available
            self.max_ref_mag_hi = None

    def _setMeasurementComponentMapping(self, specification, num_components):
        '''
        Construct the measurement-component mapping matrix from the specification.

        If specification is None, it is assumed that all measurements contain
        num_components components. Otherwise, specification should be a list of
        num_measurements tuples, where each tuples contains the (zero-based)
        indices of the components that measurement contains.

        If specification is given, the reference band must contain all components.
        '''
        num_measurements = self.responses.filters.num_filters
        if specification is None:
            self.measurement_component_mapping = np.ones((num_components, num_measurements))
            self.redshifts_exchangeable = True
        else:
            measurement_component_mapping = np.zeros((num_components, num_measurements))
            for m in range(num_measurements):
                measurement_component_mapping[specification[m], m] = 1.

            if not np.all(measurement_component_mapping[:, self.config.ref_band] == 1.):
                #TODO: Currently enforcing the ref band to have all components. This is needed
                # to be able to specifiy the fractions (IS IT??). Also ref band is currently used in the priors,
                # though the magnitudes going to the priors either have to be in the reference band
                # *OR* on their own, in which case no separation in necessary (like original BPZ case)
                raise ValueError('The reference band must contain all components.')

            #Set the mapping
            self.measurement_component_mapping = measurement_component_mapping
            #Set whether the redshifts are exchangable and so need sorting condition
            #Only need to check if there's more than one component
            if num_components > 1:
                self.redshifts_exchangeable = np.all(measurement_component_mapping[1:, :] ==
                                                     measurement_component_mapping[:-1, :])
            else:
                self.redshifts_exchangeable = None

    def _obeyPriorConditions(self, redshifts, magnitudes, ref_mag_hi):
        '''Check that the (arrays of) redshift and magnitude have sensible
        values (within bounds set by config) and that they obey the sorting conditions.
        '''
        num_components = len(redshifts)
        redshifts_in_bounds = np.all(redshifts >= 0.) \
                              and np.all(redshifts >= self.config.z_lo) \
                              and np.all(redshifts <= self.config.z_hi)
        magnitudes_in_bounds = np.all(magnitudes >= self.config.ref_mag_lo) \
                               and np.all(magnitudes <= ref_mag_hi)
        bounds_check = redshifts_in_bounds and magnitudes_in_bounds
        #Only need sorting condition if redshifts are exchangable
        # (depends on measurement_component_mapping) and if there's
        # multiple components
        if num_components>1 and self.redshifts_exchangeable:
            if self.config.sort_redshifts:
                sort_condition = np.all(redshifts[1:] >= redshifts[:-1])
            else:
                sort_condition = np.all(magnitudes[1:] >= magnitudes[:-1])
            prior_checks_okay = sort_condition and bounds_check
        else:
            prior_checks_okay = bounds_check
        return prior_checks_okay

    def _lnTotalPrior(self, params):
        ''' Total prior given array of parameters.

        This returns

        p({z}, {t}, {m0}) = [1 + xi({z})] Prod_a  Lambda_a P(z_a | t_z, m_0a) P(t_a | m_0a) P(m_0a)

        i.e., it EXCLUDES the selection effect, and is NOT marginalised over template.

        Parameter array is given in order
        [z1, z2... t1_continuous, t2_continuous... m0_1, m0_2...] where tn_continuous
        is a continuous parameter 0:num_templates that gets rounded to an int
        for discrete template.
        '''
        num_components = int(len(params) // 3)

        redshifts = params[:num_components]
        templates_cont = params[num_components:2*num_components]
        magnitudes = params[2*num_components:]
        templates_disc = np.around(templates_cont).astype(int)

        #Prior conditions
        template_positive = np.all(templates_disc >= 0.)
        template_within_bounds = np.all(templates_disc <= self.responses.templates.num_templates - 1)
        template_okay = template_positive and template_within_bounds
        # ref_mag_hi must be set in config, which is okay since this function is
        # only used in SimulatedPhotometry which assumes that must be set
        redshift_magnitude_okay = self._obeyPriorConditions(redshifts, magnitudes, self.config.ref_mag_hi)

        if not (template_okay and redshift_magnitude_okay):
            return -np.inf
        else:
            lnPrior = np.log(1. + self.correlationFunction(redshifts))
            for a in range(num_components):
                tmp_ind_a = self.responses.templates.tmp_ind_to_type_ind[int(templates_disc[a])]
                lnPrior += self.lnPrior(redshifts[a], magnitudes[a])[tmp_ind_a]

            return lnPrior

    def comovingSeparation(self, z_lo, z_hi):
        '''Returns the comoving distance between two objects along the
        line of sight, given their redshifts.
        '''

        integral = quad(lambda zp: 1./np.sqrt((self.config.omega_mat * (1+zp)**3.) +
                                              (self.config.omega_k * (1+zp)**2.) +
                                              self.config.omega_lam), z_lo, z_hi)[0]
        return (3.e5 / self.config.hubble) * integral

    def lnSelection(self, flux, galaxy):
        flim = 10.**(-0.4 * galaxy.magnitude_limit)
        sigma = galaxy.ref_flux_sigma
        selection = 0.5 - (0.5 * erf((flim - flux) / (sigma * np.sqrt(2))))
        return np.log(selection)

    def lnPriorCalibrationPrior(self):
        '''Returns the prior on the prior parameters for the calibration procedure.'''
        #Default to P(theta) = 1 --> ln P(theta) = 0
        return 0.

    def correlationFunction(self, redshifts):
        #Default to no correlation
        return 0.

    @abc.abstractmethod
    def lnPrior(self, redshift, magnitude):
        pass
